/*
Copyright (c) 2019 Gabriel Ebner. All rights reserved.
Released under Apache 2.0 license as described in the file LICENSE.

Author: Gabriel Ebner
*/

namespace lean {

void initialize_feature_search();
void finalize_feature_search();

}
