// Copyright (c) 2015 Microsoft Corporation. All rights reserved.
// Released under Apache 2.0 license as described in the file LICENSE.
// DO NOT EDIT, automatically generated file, generator scripts/gen_constants_cpp.py
#include "util/name.h"
namespace lean {
void initialize_constants();
void finalize_constants();
name const & get_absurd_name();
name const & get_acc_cases_on_name();
name const & get_acc_rec_name();
name const & get_and_name();
name const & get_and_cases_on_name();
name const & get_and_elim_left_name();
name const & get_and_elim_right_name();
name const & get_and_intro_name();
name const & get_and_rec_name();
name const & get_auto_param_name();
name const & get_bin_tree_empty_name();
name const & get_bin_tree_leaf_name();
name const & get_bin_tree_node_name();
name const & get_bit0_name();
name const & get_bit1_name();
name const & get_bool_name();
name const & get_bool_ff_name();
name const & get_bool_tt_name();
name const & get_cast_name();
name const & get_cast_heq_name();
name const & get_char_name();
name const & get_char_mk_name();
name const & get_char_ne_of_vne_name();
name const & get_char_of_nat_name();
name const & get_char_of_nat_ne_of_ne_name();
name const & get_coe_name();
name const & get_coe_fn_name();
name const & get_coe_sort_name();
name const & get_coe_to_lift_name();
name const & get_combinator_K_name();
name const & get_congr_name();
name const & get_congr_arg_name();
name const & get_congr_fun_name();
name const & get_decidable_name();
name const & get_decidable_to_bool_name();
name const & get_dite_name();
name const & get_empty_name();
name const & get_eq_name();
name const & get_eq_cases_on_name();
name const & get_eq_drec_name();
name const & get_eq_mp_name();
name const & get_eq_mpr_name();
name const & get_eq_rec_name();
name const & get_eq_refl_name();
name const & get_eq_subst_name();
name const & get_eq_symm_name();
name const & get_eq_trans_name();
name const & get_eq_false_intro_name();
name const & get_eq_of_heq_name();
name const & get_eq_rec_heq_name();
name const & get_eq_self_iff_true_name();
name const & get_eq_true_intro_name();
name const & get_Exists_name();
name const & get_expr_name();
name const & get_expr_subst_name();
name const & get_false_name();
name const & get_false_rec_name();
name const & get_false_of_true_eq_false_name();
name const & get_fin_mk_name();
name const & get_fin_ne_of_vne_name();
name const & get_forall_congr_name();
name const & get_forall_congr_eq_name();
name const & get_forall_not_of_not_exists_name();
name const & get_format_name();
name const & get_function_name();
name const & get_funext_name();
name const & get_has_add_name();
name const & get_has_add_add_name();
name const & get_has_andthen_andthen_name();
name const & get_has_bind_and_then_name();
name const & get_has_bind_seq_name();
name const & get_has_coe_t_name();
name const & get_has_coe_to_fun_name();
name const & get_has_coe_to_sort_name();
name const & get_has_div_div_name();
name const & get_has_emptyc_emptyc_name();
name const & get_has_insert_insert_name();
name const & get_has_neg_neg_name();
name const & get_has_one_name();
name const & get_has_one_one_name();
name const & get_has_orelse_orelse_name();
name const & get_has_repr_name();
name const & get_has_sep_sep_name();
name const & get_has_singleton_singleton_name();
name const & get_has_sizeof_name();
name const & get_has_sizeof_mk_name();
name const & get_has_sub_sub_name();
name const & get_has_to_format_name();
name const & get_has_well_founded_name();
name const & get_has_well_founded_r_name();
name const & get_has_well_founded_wf_name();
name const & get_has_zero_name();
name const & get_has_zero_zero_name();
name const & get_heq_name();
name const & get_heq_refl_name();
name const & get_heq_symm_name();
name const & get_heq_trans_name();
name const & get_heq_of_eq_name();
name const & get_hole_command_name();
name const & get_id_name();
name const & get_id_delta_name();
name const & get_id_rhs_name();
name const & get_if_neg_name();
name const & get_if_pos_name();
name const & get_iff_name();
name const & get_iff_intro_name();
name const & get_iff_mp_name();
name const & get_iff_mpr_name();
name const & get_iff_refl_name();
name const & get_iff_rfl_name();
name const & get_iff_symm_name();
name const & get_iff_trans_name();
name const & get_iff_false_intro_name();
name const & get_iff_true_intro_name();
name const & get_imp_congr_name();
name const & get_imp_congr_ctx_name();
name const & get_imp_congr_ctx_eq_name();
name const & get_imp_congr_eq_name();
name const & get_implies_name();
name const & get_implies_of_if_neg_name();
name const & get_implies_of_if_pos_name();
name const & get_int_name();
name const & get_int_bit0_nonneg_name();
name const & get_int_bit0_pos_name();
name const & get_int_bit1_nonneg_name();
name const & get_int_bit1_pos_name();
name const & get_int_nat_abs_bit0_step_name();
name const & get_int_nat_abs_bit1_nonneg_step_name();
name const & get_int_nat_abs_one_name();
name const & get_int_nat_abs_zero_name();
name const & get_int_ne_neg_of_ne_name();
name const & get_int_ne_neg_of_pos_name();
name const & get_int_ne_of_nat_ne_nonneg_case_name();
name const & get_int_neg_ne_of_pos_name();
name const & get_int_neg_ne_zero_of_ne_name();
name const & get_int_one_nonneg_name();
name const & get_int_one_pos_name();
name const & get_int_zero_ne_neg_of_ne_name();
name const & get_int_zero_nonneg_name();
name const & get_interactive_executor_name();
name const & get_interactive_param_desc_name();
name const & get_interactive_parse_name();
name const & get_io_name();
name const & get_io_core_name();
name const & get_io_rand_nat_name();
name const & get_is_associative_name();
name const & get_is_associative_assoc_name();
name const & get_is_commutative_name();
name const & get_is_commutative_comm_name();
name const & get_is_valid_char_range_1_name();
name const & get_is_valid_char_range_2_name();
name const & get_ite_name();
name const & get_lean_parser_name();
name const & get_lean_parser_pexpr_name();
name const & get_lean_parser_reflectable_expr_name();
name const & get_lean_parser_tk_name();
name const & get_left_comm_name();
name const & get_list_name();
name const & get_list_cons_name();
name const & get_list_nil_name();
name const & get_match_failed_name();
name const & get_monad_name();
name const & get_monad_fail_name();
name const & get_monad_from_pure_bind_name();
name const & get_monad_io_environment_impl_name();
name const & get_monad_io_file_system_impl_name();
name const & get_monad_io_impl_name();
name const & get_monad_io_net_system_impl_name();
name const & get_monad_io_process_impl_name();
name const & get_monad_io_random_impl_name();
name const & get_monad_io_terminal_impl_name();
name const & get_name_anonymous_name();
name const & get_name_mk_numeral_name();
name const & get_name_mk_string_name();
name const & get_nat_name();
name const & get_nat_add_name();
name const & get_nat_bit0_lt_name();
name const & get_nat_bit0_lt_bit1_name();
name const & get_nat_bit0_ne_name();
name const & get_nat_bit0_ne_bit1_name();
name const & get_nat_bit0_ne_one_name();
name const & get_nat_bit0_ne_zero_name();
name const & get_nat_bit1_lt_name();
name const & get_nat_bit1_lt_bit0_name();
name const & get_nat_bit1_ne_name();
name const & get_nat_bit1_ne_bit0_name();
name const & get_nat_bit1_ne_one_name();
name const & get_nat_bit1_ne_zero_name();
name const & get_nat_cases_on_name();
name const & get_nat_has_add_name();
name const & get_nat_has_one_name();
name const & get_nat_has_zero_name();
name const & get_nat_le_of_lt_name();
name const & get_nat_le_refl_name();
name const & get_nat_one_lt_bit0_name();
name const & get_nat_one_lt_bit1_name();
name const & get_nat_one_ne_bit0_name();
name const & get_nat_one_ne_bit1_name();
name const & get_nat_one_ne_zero_name();
name const & get_nat_succ_name();
name const & get_nat_zero_name();
name const & get_nat_zero_lt_bit0_name();
name const & get_nat_zero_lt_bit1_name();
name const & get_nat_zero_lt_one_name();
name const & get_nat_zero_ne_bit0_name();
name const & get_nat_zero_ne_bit1_name();
name const & get_nat_zero_ne_one_name();
name const & get_ne_name();
name const & get_neq_of_not_iff_name();
name const & get_not_name();
name const & get_not_of_eq_false_name();
name const & get_of_eq_true_name();
name const & get_opt_param_name();
name const & get_or_name();
name const & get_out_param_name();
name const & get_pprod_name();
name const & get_pprod_fst_name();
name const & get_pprod_mk_name();
name const & get_pprod_snd_name();
name const & get_prod_mk_name();
name const & get_propext_name();
name const & get_psigma_name();
name const & get_psigma_cases_on_name();
name const & get_psigma_fst_name();
name const & get_psigma_mk_name();
name const & get_psigma_snd_name();
name const & get_psum_name();
name const & get_psum_cases_on_name();
name const & get_psum_inl_name();
name const & get_psum_inr_name();
name const & get_punit_name();
name const & get_punit_cases_on_name();
name const & get_punit_star_name();
name const & get_quot_lift_name();
name const & get_quot_mk_name();
name const & get_reflected_name();
name const & get_reflected_subst_name();
name const & get_repr_name();
name const & get_rfl_name();
name const & get_scope_trace_name();
name const & get_set_of_name();
name const & get_sizeof_name();
name const & get_string_name();
name const & get_string_empty_name();
name const & get_string_empty_ne_str_name();
name const & get_string_str_name();
name const & get_string_str_ne_empty_name();
name const & get_string_str_ne_str_left_name();
name const & get_string_str_ne_str_right_name();
name const & get_subsingleton_name();
name const & get_subsingleton_elim_name();
name const & get_subsingleton_helim_name();
name const & get_subtype_name();
name const & get_subtype_mk_name();
name const & get_subtype_rec_name();
name const & get_subtype_val_name();
name const & get_tactic_name();
name const & get_tactic_mk_inj_eq_name();
name const & get_tactic_triv_name();
name const & get_tactic_try_name();
name const & get_thunk_name();
name const & get_to_fmt_name();
name const & get_to_pexpr_name();
name const & get_trans_rel_left_name();
name const & get_trans_rel_right_name();
name const & get_true_name();
name const & get_true_intro_name();
name const & get_true_eq_false_of_false_name();
name const & get_unification_hint_name();
name const & get_unification_hint_mk_name();
name const & get_unit_name();
name const & get_unit_star_name();
name const & get_user_attribute_name();
name const & get_user_attribute_parse_reflect_name();
name const & get_vm_monitor_name();
name const & get_well_founded_fix_name();
name const & get_well_founded_fix_eq_name();
name const & get_well_founded_tactics_name();
name const & get_well_founded_tactics_dec_tac_name();
name const & get_well_founded_tactics_default_name();
name const & get_well_founded_tactics_rel_tac_name();
name const & get_widget_term_goal_widget_name();
}
